@@ -2,12 +2,13 @@
  * mpeg_system.c: TS, PS and PES management
  *****************************************************************************
  * Copyright (C) 1998, 1999, 2000 VideoLAN
- * $Id: mpeg_system.c,v 1.37 2001/02/19 19:08:59 massiot Exp $
+ * $Id: mpeg_system.c,v 1.50 2001/04/13 01:49:22 henri Exp $
  *
  * Authors: Christophe Massiot <massiot@via.ecp.fr>
  *          Michel Lespinasse <walken@via.ecp.fr>
  *          Benot Steiner <benny@via.ecp.fr>
  *          Samuel Hocevar <sam@via.ecp.fr>
+ *          Henri Fallon <henri@via.ecp.fr>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,6 +31,7 @@
 #include "defs.h"
 
 #include <stdlib.h>
+#include <string.h>                                    /* memcpy(), memset() */
 
 #include "config.h"
 #include "common.h"
@@ -51,6 +53,8 @@
  * Local prototypes
  *****************************************************************************/
 
+static void input_DecodePAT( input_thread_t *, es_descriptor_t *);
+static void input_DecodePMT( input_thread_t *, es_descriptor_t *);
 
 /*
  * PES Packet management
@@ -61,6 +65,7 @@
  *****************************************************************************
  * Small utility function used to parse discontinuous headers safely. Copies
  * i_buf_len bytes of data to a buffer and returns the size copied.
+ * It also solves some alignment problems on non-IA-32, non-PPC processors.
  * This is a variation on the theme of input_ext-dec.h:GetChunk().
  *****************************************************************************/
 static __inline__ size_t MoveChunk( byte_t * p_dest,
@@ -220,8 +225,10 @@ void input_ParsePES( input_thread_t * p_input, es_descriptor_t * p_es )
                     }
                     p_pes->i_pts = input_ClockGetTS( p_input, p_es->p_pgrm,
                     ( ((mtime_t)(p_full_header[2] & 0x0E) << 29) |
-                      (((mtime_t)U16_AT(p_full_header + 3) << 14) - (1 << 14)) |
-                      ((mtime_t)U16_AT(p_full_header + 5) >> 1) ) );
+                      ((mtime_t)(p_full_header[3]) << 22) |
+                      ((mtime_t)(p_full_header[4] & 0xFE) << 14) |
+                      ((mtime_t)p_full_header[5] << 7) |
+                      ((mtime_t)p_full_header[6] >> 1) ) );
 
                     if( b_has_dts )
                     {
@@ -253,7 +260,7 @@ void input_ParsePES( input_thread_t * p_input, es_descriptor_t * p_es )
                 /* Cannot fail because the previous one succeeded. */
                 MoveChunk( NULL, &p_data, &p_byte, 6 );
 
-                while( *p_byte == 0xFF && i_pes_header_size < 22 )
+                while( *p_byte == 0xFF && i_pes_header_size < 23 )
                 {
                     i_pes_header_size++;
                     if( MoveChunk( NULL, &p_data, &p_byte, 1 ) != 1 )
@@ -265,7 +272,7 @@ void input_ParsePES( input_thread_t * p_input, es_descriptor_t * p_es )
                         return;
                     }
                 }
-                if( i_pes_header_size == 22 )
+                if( i_pes_header_size == 23 )
                 {
                     intf_ErrMsg( "Too much MPEG-1 stuffing" );
                     p_input->pf_delete_pes( p_input->p_method_data, p_pes );
@@ -310,9 +317,10 @@ void input_ParsePES( input_thread_t * p_input, es_descriptor_t * p_es )
                     }
 
                     p_pes->i_pts = input_ClockGetTS( p_input, p_es->p_pgrm,
-                      ( ((mtime_t)(p_ts[0] & 0x0E) << 29) |
-                        (((mtime_t)U16_AT(p_ts + 1) << 14) - (1 << 14)) |
-                        ((mtime_t)U16_AT(p_ts + 3) >> 1) ) );
+                       ( ((mtime_t)(p_ts[0] & 0x0E) << 29) |
+                         (((mtime_t)U32_AT(p_ts) & 0xFFFE00) << 6) |
+                         ((mtime_t)p_ts[3] << 7) |
+                         ((mtime_t)p_ts[4] >> 1) ) );
 
                     if( b_has_dts )
                     {
@@ -330,8 +338,9 @@ void input_ParsePES( input_thread_t * p_input, es_descriptor_t * p_es )
                         p_pes->i_dts = input_ClockGetTS( p_input,
                                                          p_es->p_pgrm,
                             ( ((mtime_t)(p_ts[0] & 0x0E) << 29) |
-                              (((mtime_t)U16_AT(p_ts + 1) << 14) - (1 << 14)) |
-                              ((mtime_t)U16_AT(p_ts + 3) >> 1) ) );
+                              (((mtime_t)U32_AT(p_ts) & 0xFFFE00) << 6) |
+                              ((mtime_t)p_ts[3] << 7) |
+                              ((mtime_t)p_ts[4] >> 1) ) );
                     }
                 }
             }
@@ -497,17 +506,21 @@ static u16 GetID( data_packet_t * p_data )
 {
     u16         i_id;
 
-    i_id = p_data->p_buffer[3];                                 /* stream_id */
+    i_id = p_data->p_payload_start[3];                                 /* stream_id */
     if( i_id == 0xBD )
     {
+        /* FIXME : this is not valid if the header is split in multiple
+         * packets */
         /* stream_private_id */
-        i_id |= p_data->p_buffer[ 9 + p_data->p_buffer[8] ] << 8;
+        i_id |= p_data->p_payload_start[ 9 + p_data->p_payload_start[8] ] << 8;
     }
     return( i_id );
 }
 
 /*****************************************************************************
  * DecodePSM: Decode the Program Stream Map information
+ *****************************************************************************
+ * FIXME : loads are not aligned in this function
  *****************************************************************************/
 static void DecodePSM( input_thread_t * p_input, data_packet_t * p_data )
 {
@@ -525,7 +538,7 @@ static void DecodePSM( input_thread_t * p_input, data_packet_t * p_data )
     }
 
     if( p_demux->b_has_PSM
-        && p_demux->i_PSM_version == (p_data->p_buffer[6] & 0x1F) )
+        && p_demux->i_PSM_version == (p_data->p_payload_start[6] & 0x1F) )
     {
         /* Already got that one. */
         return;
@@ -533,7 +546,7 @@ static void DecodePSM( input_thread_t * p_input, data_packet_t * p_data )
 
     intf_DbgMsg( "Building PSM" );
     p_demux->b_has_PSM = 1;
-    p_demux->i_PSM_version = p_data->p_buffer[6] & 0x1F;
+    p_demux->i_PSM_version = p_data->p_payload_start[6] & 0x1F;
 
     /* Go to elementary_stream_map_length, jumping over
      * program_stream_info. */
@@ -644,8 +657,9 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
     u32                 i_code;
     es_descriptor_t *   p_es = NULL;
 
-    i_code = U32_AT( p_data->p_buffer );
-    if( i_code > 0x1BC ) /* ES start code */
+    i_code = p_data->p_payload_start[3];
+
+    if( i_code > 0xBC ) /* ES start code */
     {
         u16                 i_id;
         int                 i_dummy;
@@ -682,13 +696,14 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
                                     i_id, 0 );
                 if( p_es != NULL )
                 {
-                    p_es->i_stream_id = p_data->p_buffer[3];
+                    p_es->i_stream_id = p_data->p_payload_start[3];
 
                     /* Set stream type and auto-spawn. */
                     if( (i_id & 0xF0) == 0xE0 )
                     {
                         /* MPEG video */
                         p_es->i_type = MPEG2_VIDEO_ES;
+                        p_es->i_cat = VIDEO_ES;
 #ifdef AUTO_SPAWN
                         if( !p_input->stream.b_seekable )
                             input_SelectES( p_input, p_es );
@@ -699,6 +714,7 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
                         /* MPEG audio */
                         p_es->i_type = MPEG2_AUDIO_ES;
                         p_es->b_audio = 1;
+                        p_es->i_cat = AUDIO_ES;
 #ifdef AUTO_SPAWN
                         if( !p_input->stream.b_seekable )
                         if( main_GetIntVariable( INPUT_CHANNEL_VAR, 0 )
@@ -718,6 +734,7 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
                         /* AC3 audio (0x80->0x8F) */
                         p_es->i_type = AC3_AUDIO_ES;
                         p_es->b_audio = 1;
+                        p_es->i_cat = AUDIO_ES;
 #ifdef AUTO_SPAWN
                         if( !p_input->stream.b_seekable )
                         if( main_GetIntVariable( INPUT_CHANNEL_VAR, 0 )
@@ -736,6 +753,7 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
                     {
                         /* Subtitles video (0x20->0x3F) */
                         p_es->i_type = DVD_SPU_ES;
+                        p_es->i_cat = SPU_ES;
 #ifdef AUTO_SPAWN
                         if( main_GetIntVariable( INPUT_SUBTITLE_VAR, -1 )
                                 == ((p_es->i_id & 0x1F00) >> 8) )
@@ -750,6 +768,7 @@ es_descriptor_t * input_ParsePS( input_thread_t * p_input,
                         /* LPCM audio (0xA0->0xAF) */
                         p_es->i_type = LPCM_AUDIO_ES;
                         p_es->b_audio = 1;
+                        p_es->i_cat = AUDIO_ES;
                         /* FIXME : write the decoder */
                     }
                     else
@@ -774,7 +793,7 @@ void input_DemuxPS( input_thread_t * p_input, data_packet_t * p_data )
     boolean_t           b_trash = 0;
     es_descriptor_t *   p_es = NULL;
 
-    i_code = U32_AT( p_data->p_buffer );
+    i_code = U32_AT( p_data->p_payload_start );
     if( i_code <= 0x1BC )
     {
         switch( i_code )
@@ -785,30 +804,52 @@ void input_DemuxPS( input_thread_t * p_input, data_packet_t * p_data )
                 mtime_t         scr_time;
                 u32             i_mux_rate;
 
-                if( (p_data->p_buffer[4] & 0xC0) == 0x40 )
+                if( (p_data->p_payload_start[4] & 0xC0) == 0x40 )
                 {
                     /* MPEG-2 */
+                    byte_t      p_header[14];
+                    byte_t *    p_byte;
+                    p_byte = p_data->p_payload_start;
+
+                    if( MoveChunk( p_header, &p_data, &p_byte, 14 ) != 14 )
+                    {
+                        intf_WarnMsg( 3, "Packet too short to have a header" );
+                        b_trash = 1;
+                        break;
+                    }
                     scr_time =
-                         ((mtime_t)(p_data->p_buffer[4] & 0x38) << 27) |
-                         ((mtime_t)(U32_AT(p_data->p_buffer + 4) & 0x03FFF800)
+                         ((mtime_t)(p_header[4] & 0x38) << 27) |
+                         ((mtime_t)(U32_AT(p_header + 4) & 0x03FFF800)
                                         << 4) |
-                         ((mtime_t)(U32_AT(p_data->p_buffer + 6) & 0x03FFF800)
+                         ((( ((mtime_t)U16_AT(p_header + 6) << 16)
+                            | (mtime_t)U16_AT(p_header + 8) ) & 0x03FFF800)
                                         >> 11);
 
                     /* mux_rate */
-                    i_mux_rate = (U32_AT(p_data->p_buffer + 10) & 0xFFFFFC00);
+                    i_mux_rate = ((u32)U16_AT(p_header + 10) << 6)
+                                   | (p_header[12] >> 2);
                 }
                 else
                 {
                     /* MPEG-1 SCR is like PTS. */
+                    byte_t      p_header[12];
+                    byte_t *    p_byte;
+                    p_byte = p_data->p_payload_start;
+
+                    if( MoveChunk( p_header, &p_data, &p_byte, 12 ) != 12 )
+                    {
+                        intf_WarnMsg( 3, "Packet too short to have a header" );
+                        b_trash = 1;
+                        break;
+                    }
                     scr_time =
-                         ((mtime_t)(p_data->p_buffer[4] & 0x0E) << 29) |
-                         (((mtime_t)U16_AT(p_data->p_buffer + 5) << 14)
-                           - (1 << 14)) |
-                         ((mtime_t)U16_AT(p_data->p_buffer + 7) >> 1);
+                         ((mtime_t)(p_header[4] & 0x0E) << 29) |
+                         (((mtime_t)U32_AT(p_header + 4) & 0xFFFE00) << 6) |
+                         ((mtime_t)p_header[7] << 7) |
+                         ((mtime_t)p_header[8] >> 1);
 
                     /* mux_rate */
-                    i_mux_rate = (U32_AT(p_data->p_buffer + 8) & 0x8FFFFE);
+                    i_mux_rate = (U32_AT(p_header + 8) & 0x7FFFFE) >> 1;
                 }
                 /* Call the pace control. */
                 input_ClockManageRef( p_input, p_input->stream.pp_programs[0],
@@ -842,7 +883,7 @@ void input_DemuxPS( input_thread_t * p_input, data_packet_t * p_data )
         default:
             /* This should not happen */
             b_trash = 1;
-            intf_WarnMsg( 1, "Unwanted packet received with start code %x",
+            intf_WarnMsg( 1, "Unwanted packet received with start code 0x%.8x",
                           i_code );
         }
     }
@@ -887,21 +928,19 @@ void input_DemuxPS( input_thread_t * p_input, data_packet_t * p_data )
  *****************************************************************************/
 void input_DemuxTS( input_thread_t * p_input, data_packet_t * p_data )
 {
-    int                 i_pid, i_dummy;
+    u16                 i_pid;
+    int                 i_dummy;
     boolean_t           b_adaptation;         /* Adaptation field is present */
     boolean_t           b_payload;                 /* Packet carries payload */
     boolean_t           b_unit_start;  /* A PSI or a PES start in the packet */
     boolean_t           b_trash = 0;             /* Is the packet unuseful ? */
     boolean_t           b_lost = 0;             /* Was there a packet loss ? */
+    boolean_t           b_psi = 0;                        /* Is this a PSI ? */
     es_descriptor_t *   p_es = NULL;
     es_ts_data_t *      p_es_demux = NULL;
     pgrm_ts_data_t *    p_pgrm_demux = NULL;
 
     #define p (p_data->p_buffer)
-
-    //intf_DbgMsg("input debug: TS-demultiplexing packet %p, pid %d",
-    //            p_ts_packet, U16_AT(&p[1]) & 0x1fff);
-
     /* Extract flags values from TS common header. */
     i_pid = U16_AT(&p[1]) & 0x1fff;
     b_unit_start = (p[1] & 0x40);
@@ -910,42 +949,37 @@ void input_DemuxTS( input_thread_t * p_input, data_packet_t * p_data )
 
     /* Find out the elementary stream. */
     vlc_mutex_lock( &p_input->stream.stream_lock );
-
-// kludge    
-if ( i_pid == 0x78 )
-{
-    p_es = input_FindES( p_input, 0x78 );
-    p_es->i_type = MPEG2_VIDEO_ES;
-    if( p_es->p_pes == NULL )
-      intf_ErrMsg("Got p_es . p_es->p_pes == null ? %d",p_es->p_pes == NULL);
-}
-else
-{
-    p_es = NULL;
-    b_trash = 1;
-}
-
+        
+    p_es= input_FindES( p_input, i_pid );
     
+    if( (p_es != NULL) && (p_es->p_demux_data != NULL) )
+    {
+        p_es_demux = (es_ts_data_t *)p_es->p_demux_data;
+        
+        if( p_es_demux->b_psi )
+            b_psi = 1;
+        else
+            p_pgrm_demux = (pgrm_ts_data_t *)p_es->p_pgrm->p_demux_data; 
+    }
+
     vlc_mutex_lock( &p_input->stream.control.control_lock );
-    if( p_es == NULL || p_es->p_decoder_fifo == NULL
-         || (p_es->b_audio && p_input->stream.control.b_mute) )
+    if( ( p_es == NULL ) || (p_es->b_audio && p_input->stream.control.b_mute) )
     {
         /* Not selected. Just read the adaptation field for a PCR. */
         b_trash = 1;
     }
+    else if( p_es->p_decoder_fifo == NULL  && !b_psi )
+      b_trash =1; 
+
     vlc_mutex_unlock( &p_input->stream.control.control_lock );
     vlc_mutex_unlock( &p_input->stream.stream_lock );
 
-// kludge
-    if (p_es != NULL )
-    {
-    
-    p_es_demux = (es_ts_data_t *)p_es->p_demux_data;
-    p_pgrm_demux = (pgrm_ts_data_t *)p_es->p_pgrm->p_demux_data; 
-    
-    p_pgrm_demux->i_pcr_pid = 0x78;
 
-    if( (p_es->p_decoder_fifo != NULL) || (p_pgrm_demux->i_pcr_pid == i_pid) )
+    /* Don't change the order of the tests : if b_psi then p_pgrm_demux 
+     * may still be null. Who said it was ugly ? */
+    if( ( p_es != NULL ) && 
+        ((p_es->p_decoder_fifo != NULL) || b_psi 
+                                   || (p_pgrm_demux->i_pcr_pid == i_pid) ) )
     {
 #ifdef STATS
         p_es->c_packets++;
@@ -1082,13 +1116,10 @@ else
             } /* not continuous */
         } /* continuity */
     } /* if selected or PCR */
-
-    }
     
     /* Trash the packet if it has no payload or if it isn't selected */
     if( b_trash )
     {
-        
         p_input->pf_delete_packet( p_input->p_method_data, p_data );
 #ifdef STATS
         p_input->c_packets_trashed++;
@@ -1096,24 +1127,403 @@ else
     }
     else
     {
-        if( p_es_demux->b_psi )
+        if( b_psi )
         {
-//debug
-//printf("DemuxTS : Was a PSI\n");
             /* The payload contains PSI tables */
-#if 0
-            /* FIXME ! write the PSI decoder :p */
             input_DemuxPSI( p_input, p_data, p_es,
                             b_unit_start, b_lost );
-#endif
+
         }
         else
         {
             /* The payload carries a PES stream */
             input_GatherPES( p_input, p_data, p_es, b_unit_start, b_lost ); 
         }
+
     }
 
 #undef p
 
 }
+
+/*
+ * PSI demultiplexing and decoding
+ */
+
+/*****************************************************************************
+ * DemuxPSI : makes up complete PSI data
+ *****************************************************************************/
+void input_DemuxPSI( input_thread_t * p_input, data_packet_t * p_data, 
+        es_descriptor_t * p_es, boolean_t b_unit_start, boolean_t b_lost )
+{
+    es_ts_data_t  * p_demux_data;
+    
+    p_demux_data = (es_ts_data_t *)p_es->p_demux_data;
+
+#define p_psi (p_demux_data->p_psi_section)
+#define p (p_data->p_payload_start)
+
+    if( b_unit_start )
+    {
+        /* unit_start set to 1 -> presence of a pointer field
+         * (see ISO/IEC 13818 (2.4.4.2) which should be set to 0x00 */
+        if( (u8)p[0] != 0x00 )
+        {
+            intf_WarnMsg( 2, 
+                          "Non zero pointer field found. Trying to continue" );
+            p+=(u8)p[0];
+        }
+        else
+            p++;
+
+        /* This is the begining of a new section */
+
+        if( ((u8)(p[1]) & 0xc0) != 0x80 ) 
+        {
+            intf_WarnMsg( 2, "Invalid PSI packet" );
+            p_psi->b_trash = 1;
+        }
+        else 
+        {
+            p_psi->i_section_length = U16_AT(p+1) & 0x0fff;
+            p_psi->b_section_complete = 0;
+            p_psi->i_read_in_section = 0;
+            p_psi->i_section_number = (u8)p[6];
+
+            if( p_psi->b_is_complete || p_psi->i_section_number == 0 )
+            {
+                /* This is a new PSI packet */
+                p_psi->b_is_complete = 0;
+                p_psi->b_trash = 0;
+                p_psi->i_version_number = ( p[5] >> 1 ) & 0x1f;
+                p_psi->i_last_section_number = (u8)p[7];
+
+                /* We'll write at the begining of the buffer */
+                p_psi->p_current = p_psi->buffer;
+            }
+            else
+            {
+                if( p_psi->b_section_complete )
+                {
+                    /* New Section of an already started PSI */
+                    p_psi->b_section_complete = 0;
+                    
+                    if( p_psi->i_version_number != (( p[5] >> 1 ) & 0x1f) )
+                    {
+                        intf_WarnMsg( 2,"PSI version differs inside same PAT" );
+                        p_psi->b_trash = 1;
+                    }
+                    if( p_psi->i_section_number + 1 != (u8)p[6] )
+                    {
+                        intf_WarnMsg( 2, 
+                                "PSI Section discontinuity. Packet lost ?");
+                        p_psi->b_trash = 1;
+                    }
+                    else
+                        p_psi->i_section_number++;
+                }
+                else
+                {
+                    intf_WarnMsg( 2, "Received unexpected new PSI section" );
+                    p_psi->b_trash = 1;
+                }
+            }
+        }
+    } /* b_unit_start */
+    
+    if( !p_psi->b_trash )
+    {
+        /* read */
+        if( (p_data->p_payload_end - p) >=
+            ( p_psi->i_section_length - p_psi->i_read_in_section ) )
+        {
+            /* The end of the section is in this TS packet */
+            memcpy( p_psi->p_current, p, 
+            (p_psi->i_section_length - p_psi->i_read_in_section) );
+    
+            p_psi->b_section_complete = 1;
+            p_psi->p_current += 
+                (p_psi->i_section_length - p_psi->i_read_in_section);
+                        
+            if( p_psi->i_section_number == p_psi->i_last_section_number )
+            {
+                /* This was the last section of PSI */
+                p_psi->b_is_complete = 1;
+            }
+        }
+        else
+        {
+            memcpy( p_psi->buffer, p, p_data->p_payload_end - p );
+            p_psi->i_read_in_section+= p_data->p_payload_end - p;
+
+            p_psi->p_current += p_data->p_payload_end - p;
+        }
+    }
+
+    if ( p_psi->b_is_complete )
+    {
+        switch( p_demux_data->i_psi_type)
+        {
+            case PSI_IS_PAT:
+                input_DecodePAT( p_input, p_es );
+                break;
+            case PSI_IS_PMT:
+                input_DecodePMT( p_input, p_es );
+                break;
+            default:
+                intf_WarnMsg(2, "Received unknown PSI in demuxPSI");
+        }
+    }
+#undef p_psi    
+#undef p
+    
+    return ;
+}
+
+/*****************************************************************************
+ * DecodePAT : Decodes Programm association table and deal with it
+ *****************************************************************************/
+static void input_DecodePAT( input_thread_t * p_input, es_descriptor_t * p_es )
+{
+    
+    stream_ts_data_t  * p_stream_data;
+    es_ts_data_t      * p_demux_data;
+
+    p_demux_data = (es_ts_data_t *)p_es->p_demux_data;
+    p_stream_data = (stream_ts_data_t *)p_input->stream.p_demux_data;
+    
+#define p_psi (p_demux_data->p_psi_section)
+
+    if( p_stream_data->i_pat_version != p_psi->i_version_number )
+    {
+        /* PAT has changed. We are going to delete all programms and 
+         * create new ones. We chose not to only change what was needed
+         * as a PAT change may mean the stream is radically changing and
+         * this is a secure method to avoid krashes */
+        pgrm_descriptor_t * p_pgrm;
+        es_descriptor_t   * p_current_es;
+        es_ts_data_t      * p_es_demux;
+        pgrm_ts_data_t    * p_pgrm_demux;
+        byte_t            * p_current_data;           
+        
+        int                 i_section_length,i_program_id,i_pmt_pid;
+        int                 i_loop, i_current_section;
+        
+        p_current_data = p_psi->buffer;
+
+
+        for( i_loop = 0; i_loop < p_input->stream.i_pgrm_number; i_loop++ )
+        {
+            input_DelProgram( p_input, p_input->stream.pp_programs[i_loop] );
+        }
+        
+        do
+        {
+            i_section_length = U16_AT(p_current_data+1) & 0x0fff;
+            i_current_section = (u8)p_current_data[6];
+    
+            for( i_loop = 0; i_loop < (i_section_length-9)/4 ; i_loop++ )
+            {
+                i_program_id = U16_AT(p_current_data + i_loop*4 + 8);
+                i_pmt_pid = U16_AT( p_current_data + i_loop*4 + 10) & 0x1fff;
+    
+                /* If program = 0, we're having info about NIT not PMT */
+                if( i_program_id )
+                {
+                    /* Add this program */
+                    p_pgrm = input_AddProgram( p_input, i_program_id, 
+                                               sizeof( pgrm_ts_data_t ) );
+                   
+                    /* whatis the PID of the PMT of this program */
+                    p_pgrm_demux = (pgrm_ts_data_t *)p_pgrm->p_demux_data;
+                    p_pgrm_demux->i_pmt_version = PMT_UNINITIALIZED;
+    
+                    /* Add the PMT ES to this program */
+                    p_current_es = input_AddES( p_input, p_pgrm,(u16)i_pmt_pid,
+                                        sizeof( es_ts_data_t) );
+                    p_es_demux = (es_ts_data_t *)p_current_es->p_demux_data;
+                    p_es_demux->b_psi = 1;
+                    p_es_demux->i_psi_type = PSI_IS_PMT;
+                    
+                    p_es_demux->p_psi_section = 
+                                            malloc( sizeof( psi_section_t ) );
+                    p_es_demux->p_psi_section->b_is_complete = 0;
+                }
+            }
+            
+            p_current_data+=3+i_section_length;
+            
+        } while( i_current_section < p_psi->i_last_section_number );
+        
+        /* Go to the beginning of the next section*/
+        p_stream_data->i_pat_version = p_psi->i_version_number;
+
+    }
+#undef p_psi    
+
+}
+
+/*****************************************************************************
+ * DecodePMT : decode a given Program Stream Map
+ * ***************************************************************************
+ * When the PMT changes, it may mean a deep change in the stream, and it is
+ * careful to delete the ES and add them again. If the PMT doesn't change,
+ * there no need to do anything.
+ *****************************************************************************/
+static void input_DecodePMT( input_thread_t * p_input, es_descriptor_t * p_es )
+{
+
+    pgrm_ts_data_t            * p_pgrm_data;
+    es_ts_data_t              * p_demux_data;
+
+    p_demux_data = (es_ts_data_t *)p_es->p_demux_data;
+    p_pgrm_data = (pgrm_ts_data_t *)p_es->p_pgrm->p_demux_data;
+    
+#define p_psi (p_demux_data->p_psi_section)
+
+    if( p_psi->i_version_number != p_pgrm_data->i_pmt_version ) 
+    {
+        es_descriptor_t   * p_new_es;  
+        es_ts_data_t      * p_es_demux;
+        byte_t            * p_current_data, * p_current_section;
+        int                 i_section_length,i_current_section;
+        int                 i_prog_info_length, i_loop;
+        int                 i_es_info_length, i_pid, i_stream_type;
+        int                 i_audio_es, i_spu_es;
+        int                 i_required_audio_es, i_required_spu_es;
+        
+        p_current_section = p_psi->buffer;
+        p_current_data = p_psi->buffer;
+        
+        p_pgrm_data->i_pcr_pid = U16_AT(p_current_section + 8) & 0x1fff;
+        
+        i_audio_es = 0;
+        i_spu_es = 0;
+        
+        /* Lock stream information */
+        vlc_mutex_lock( &p_input->stream.stream_lock );
+
+        /* Get the number of the required audio stream */
+        if( p_main->b_audio )
+        {
+            /* Default is the first one */
+            i_required_audio_es = main_GetIntVariable( INPUT_CHANNEL_VAR, 1 );
+            if( i_required_audio_es < 0 )
+            {
+                main_PutIntVariable( INPUT_CHANNEL_VAR, 1 );
+                i_required_audio_es = 1;
+            }
+        }
+        else
+        {
+            i_required_audio_es = 0;
+        }
+
+        /* Same thing for subtitles */
+        if( p_main->b_video )
+        {
+            /* for spu, default is none */
+            i_required_spu_es = main_GetIntVariable( INPUT_SUBTITLE_VAR, 0 );
+            if( i_required_spu_es < 0 )
+            {
+                main_PutIntVariable( INPUT_SUBTITLE_VAR, 0 );
+                i_required_spu_es = 0;
+            }
+        }
+        else
+        {
+            i_required_spu_es = 0;
+        }
+        
+        /* Delete all ES in this program  except the PSI */
+        for( i_loop=0; i_loop < p_es->p_pgrm->i_es_number; i_loop++ )
+        {
+            p_es_demux = (es_ts_data_t *)
+                         p_es->p_pgrm->pp_es[i_loop]->p_demux_data;
+            if ( ! p_es_demux->b_psi )
+            input_DelES( p_input, p_es->p_pgrm->pp_es[i_loop] );
+        }
+
+        /* Then add what we received in this PMT */
+        do
+        {
+            
+            i_section_length = U16_AT(p_current_data+1) & 0x0fff;
+            i_current_section = (u8)p_current_data[6];
+            i_prog_info_length = U16_AT(p_current_data+10) & 0x0fff;
+
+            /* For the moment we ignore program descriptors */
+            p_current_data += 12+i_prog_info_length;
+    
+            /* The end of the section, before the CRC is at 
+             * p_current_section + i_section_length -1 */
+            while( p_current_data < p_current_section + i_section_length -1 )
+            {
+                i_stream_type = (int)p_current_data[0];
+                i_pid = U16_AT( p_current_data + 1 ) & 0x1fff;
+                i_es_info_length = U16_AT( p_current_data + 3 ) & 0x0fff;
+                
+                /* Add this ES to the program */
+                p_new_es = input_AddES( p_input, p_es->p_pgrm, 
+                                        (u16)i_pid, sizeof( es_ts_data_t ) );
+
+                /* Tell the decoders what kind of stream it is */
+                p_new_es->i_type = i_stream_type;
+
+                /* Tell the interface what kind of stream it is and select 
+                 * the required ones */
+                switch( i_stream_type )
+                {
+                    case MPEG1_VIDEO_ES:
+                    case MPEG2_VIDEO_ES:
+                        p_new_es->i_cat = VIDEO_ES;
+                        input_SelectES( p_input, p_new_es );
+                        break;
+                    case MPEG1_AUDIO_ES:
+                    case MPEG2_AUDIO_ES:
+                    case AC3_AUDIO_ES :
+                    case LPCM_AUDIO_ES :
+                        p_new_es->i_cat = AUDIO_ES;
+                        i_audio_es += 1;
+                        if( i_audio_es == i_required_audio_es )
+                            input_SelectES( p_input, p_new_es );
+                        break;
+                    /* Not sure this one is fully norm-compliant */
+                    case DVD_SPU_ES :
+                        p_new_es->i_cat = SPU_ES;
+                        i_spu_es += 1;
+                        if( i_spu_es == i_required_spu_es )
+                            input_SelectES( p_input, p_new_es );
+                        break;
+                    default :
+                        p_new_es->i_cat = UNKNOWN_ES;
+                        break;
+                }
+                
+                p_current_data += 5 + i_es_info_length;
+            }
+
+            /* Go to the beginning of the next section*/
+            p_current_data += 3+i_section_length;
+           
+            p_current_section+=1;
+            
+        } while( i_current_section < p_psi->i_last_section_number );
+
+        if( i_required_audio_es > i_audio_es )
+            intf_WarnMsg( 2, "TS input: Non-existing audio es required." );
+        
+        if( i_required_spu_es > i_spu_es )
+            intf_WarnMsg( 2, "TS input: Non-existing subtitles es required." );
+        
+        p_pgrm_data->i_pmt_version = p_psi->i_version_number;
+
+        /* inform interface that stream has changed */
+        p_input->stream.b_changed = 1;
+    }
+    
+#undef p_psi
+
+    /*  Remove lock */
+    vlc_mutex_unlock( &p_input->stream.stream_lock );
+}
